# Unittests for RenderPDF Parser

import unittest
from StringIO import StringIO
import xml.dom.minidom
import urllib,os.path

from reportlab.lib import colors
from reportlab.lib.enums import TA_LEFT, TA_RIGHT, TA_CENTER, TA_JUSTIFY
from reportlab.lib.units import toLength

from Products.CMFReportTool.RenderPDF import Platypus
from Products.CMFReportTool.RenderPDF.Parser import TemplateParser,DocumentParser, PIL_Image


test_template_xml_template = """
<template filename="default.pdf"
            pagesize="A4"
            landscape="0"
            showboundary="0"
            leftmargin="1cm"
            rightmargin="1cm"
            topmargin="1cm"
            bottommargin="1cm"
            allowsplitting="1">
  <pagetemplate id="Page" >
   <frame  id="content"
           nextid="content"
           x="2cm"
           y="1.5cm"
           width="17cm"
           height="25.5cm"
           leftpadding="0.1cm"
           rightpadding="0.1cm"
           toppadding="0.5cm"
           bottompadding="0.5cm"
           showBoundary="0"/>
  </pagetemplate>
</template>
"""

document_xml_template = """
<document filename="report01.pdf">
  <title tal:content="template/title">Test Document</title>
  <author>Ulrich Eck</author>
  <subject>only for testing ...</subject>
  <content>
  %s
  </content>
</document>
"""




class TestDocumentParser(unittest.TestCase):


    def setUp(self):
        template_dom = xml.dom.minidom.parseString(test_template_xml_template)
        self.template = TemplateParser(template_dom, "iso-8859-1")

    def _parse(self, content=''):
        xml_data = document_xml_template % content
        xml_dom = xml.dom.minidom.parseString(xml_data)
        return DocumentParser(xml_dom, "iso-8859-1")

    def test_paragraph(self):
        xml_fragment ="""
        <para style="test">Test</para>
        """
        doc = self._parse(xml_fragment)
        el = doc.doc.content[0]
        assert isinstance(el, Platypus.xmlParagraph)
        assert el.style == 'test'
        assert el.rawtext == 'Test'

        xml_fragment ="""
        <para style="Bullet" bullettext="X">Test</para>
        """
        doc = self._parse(xml_fragment)
        el = doc.doc.content[0]
        assert isinstance(el, Platypus.xmlParagraph)
        assert el.style == 'Bullet'
        assert el.bulletText == 'X'
        assert el.rawtext == 'Test'



    def test_prefmt(self):
        xml_fragment ="""
        <prefmt style="test" dedent="3">Test</prefmt>
        """
        doc = self._parse(xml_fragment)
        el = doc.doc.content[0]
        assert isinstance(el, Platypus.xmlPreformattedText)
        assert el.style == 'test'
        assert el.dedent == 3
        assert el.rawtext == 'Test'

    





class TestTemplateParser(unittest.TestCase):
    pass




def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite( TestDocumentParser, ),
        ))

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')
