#
#
#   CMFReportTool - Generating PDF-Skins on the Fly from CMF-Sites
#
#   Copyright (C) 2002  net-labs Systemhaus GmbH
#                                               info@net-labs.de
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with this library; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Portal report tool.
$Id: ReportTool.py,v 1.12 2004/06/15 10:23:28 okuji Exp $
"""
__version__='$Revision: 1.12 $'[11:-2]


from StringIO import StringIO
from types import UnicodeType
import xml.dom.minidom
import urllib2
import os.path
import re

import Globals
from Globals import DTMLFile, PersistentMapping
from Acquisition import aq_base,aq_parent
from DateTime import DateTime
from AccessControl import ClassSecurityInfo,getSecurityManager
from OFS.Folder import Folder

from Products.CMFCore.utils import UniqueObject, getToolByName
from Products.CMFCore import CMFCorePermissions
from Products.CMFCore.ActionProviderBase import ActionProviderBase

from Products.PageTemplates.Expressions import restrictedTraverse
from StructuredText.ST import Basic as STX_Basic
from StructuredText.StructuredText import Document as STX_Document


from RenderPDF.Parser import TemplateParser,DocumentParser
from STXParser import STXParser
from html2pml import Html2Pml


from reportlab.pdfbase import pdfmetrics
pdfmetrics.registerFont(pdfmetrics.Font('Helvetica-WinAnsi', 'Helvetica', 'WinAnsiEncoding'))
pdfmetrics.registerFont(pdfmetrics.Font('Helvetica-Bold-WinAnsi', 'Helvetica-Bold', 'WinAnsiEncoding'))
pdfmetrics.registerFont(pdfmetrics.Font('Helvetica-BoldOblique-WinAnsi', 'Helvetica-BoldOblique', 'WinAnsiEncoding'))

def Items(**kw): return kw


def modifiedOptions():
    # Remove the existing "Properties" option and add our own.
    rval = []
    pos = -1
    for o in Folder.manage_options:
        label = o.get('label', None)
        if label != 'Properties':
            rval.append(o)
    rval[1:1] = [{'label':'Properties',
                  'action':'manage_propertiesForm'}]
    return tuple(rval)



# Thanks to Kapil for sharing this method
# it's currently not used, cause FSImage doesn't have getSize
def file2string(o):
    """ transform a zope file object into a string """

    (start,end) = (0,o.getSize())
    data = o.data
    if type(data) is type(''): #StringType
        return data[start:end]
    else: # blob chain
        pos = 0
        buf = [] #infile = StringIO(data.data)
        w = buf.append
        while data is not None:
            l =  len(data.data)
            pos = pos + l
            if pos > start:
                # We are within the range
                lstart = l - (pos - start)

                if lstart < 0: lstart = 0

                # find the endpoint
                if end <= pos:
                    lend = l - (pos - end)
                    w(data[lstart:lend])
                    break

                w(data[lstart:])

            data = data.next

    return ''.join(buf)



class ZODBHandler(urllib2.BaseHandler):

    def __init__(self, context):
        self.context = context

    def zodb_open(self, req):
        path = req.get_selector()
        path = path.split('/')
        obj = restrictedTraverse(self.context,path,getSecurityManager())

        # check type and e.g. call object if script ...
        if callable(obj):
            try:
                obj = obj()
            except:
                pass

        ## for OFS.Image-like objects
        if hasattr(obj,'_original'):
            obj = obj._original._data()
        elif hasattr(obj,'data'):
            obj = obj.data
        elif hasattr(obj, '_data'):
            obj = obj._data

        return StringIO(str(obj))



class ResourceHandler:
    ''' Wrapper for ZODB, URL, File Resources'''

    def __init__(self,context=None):
        zodbhandler = ZODBHandler(context)
        self.opener = urllib2.build_opener(zodbhandler)

    def get(self, url, data=None):
        # XXX potential Unicode problem here
        url = str(url)
        if url.startswith('file:'):
            raise urllib2.URLError, 'The file: protocol is disabled for security'

        if not (url.startswith('zodb') or url.startswith('http')): 
            site = self.opener.handle_open['zodb'][0].context
            if url.startswith('/'):
                url = site.REQUEST.BASE0 + url
            else:
                url = site.REQUEST.URL2 + '/' + url
        fd = self.opener.open(url, data)
            
        return StringIO(fd.read())



class ReportTool(UniqueObject, Folder, ActionProviderBase):
    '''
    This tool is used to supply reports to a portal.
    '''

    id = 'portal_report'
    meta_type = 'CMF Report Tool'

    _actions = []
    security = ClassSecurityInfo()


    manage_options = ( { 'label' : 'Overview', 'action' : 'manage_overview' }
                     ,
                     ) + ActionProviderBase.manage_options + modifiedOptions()

    #
    #   ZMI methods
    #
    security.declareProtected( CMFCorePermissions.ManagePortal, 'manage_overview' )
    manage_overview = DTMLFile( 'dtml/explainReportTool', globals() )

    request_varname = 'portal_report'

    security.declareProtected(CMFCorePermissions.ManagePortal, 'manage_propertiesForm')
    manage_propertiesForm = DTMLFile('dtml/ReportToolProps', globals())


    security.declareProtected(CMFCorePermissions.ManagePortal, 'manage_properties')
    def manage_properties(self, request_varname='', REQUEST=None):
        '''
        Changes portal_types properties.
        '''
        self.request_varname = str(request_varname)

        # XXX useful properties:
        # default encoding
        # default context
        # default template

        if REQUEST is not None:

            return self.manage_propertiesForm(
                self, REQUEST, manage_tabs_message='Properties changed.')
    
    security.declareProtected(CMFCorePermissions.View, 'renderPDF')
    def renderPDF(self,template_xml,document_xml,encoding='iso-8859-15',context=None):
        """
            Render document using template
            
        """
        # context given or PortalObject
        context = context or self.aq_inner.aq_parent
        rhandler = ResourceHandler(context)

        # if zope gives us the xml in unicode
        # we need to encode it before it can be parsed
        if type(template_xml) is UnicodeType:
            template_xml = self._encode(template_xml, encoding)
        if type(document_xml) is UnicodeType:
            document_xml = self._encode(document_xml, encoding)
        
                
        # create the PDFTemplate from xml
        template_dom = xml.dom.minidom.parseString(template_xml)
        template_dom.encoding = encoding
        # XXX as long as reportlab works only with iso-8859-15,
        # hardcode it !
        template = TemplateParser(template_dom, 'iso-8859-15', resourceHandler=rhandler)()

        # create the PDFDocment from xml
        document_dom = xml.dom.minidom.parseString(document_xml)
        document_dom.encoding = encoding
        # XXX as long as reportlab works only with iso-8859-15,
        # hardcode it !
        document = DocumentParser(document_dom, 'iso-8859-15', resourceHandler=rhandler)

        # create the PDF itself using the document and the template
        buf = StringIO()
        document(template,buf)
        buf.seek(0)
        return buf.read()
        

    security.declareProtected(CMFCorePermissions.View, 'convertSTX')
    def convertSTX(self,structuredString,header=1,level=1):
        """
        convert STX-Text to PML
        """
        st = STX_Basic(structuredString)
        doc = STX_Document(st)
        return STXParser(doc,header=header,level=level)
        

    security.declareProtected(CMFCorePermissions.View, 'convertHTML')
    def convertHTML(self, htmlString, encoding='iso-8859-15', isXHTML=False):
        """
        Convert simple HTML Structures to PML
        """
        h2p = Html2Pml()
        if type(htmlString) != UnicodeType:
            htmlString = unicode(htmlString, encoding)
        return h2p.convert(htmlString, isXHTML=isXHTML)



    def _encode(self, text, encoding=None):
        if encoding is not None:
            return text.encode(encoding)
        return str(text)
        
    def getStringType(self, str):
        from types import StringType, UnicodeType
        if isinstance(str, StringType):
            return 'bytes'
        elif isinstance(str, UnicodeType):
            return 'unicode'
        else:
            return 'else'
    
        

Globals.InitializeClass(ReportTool)
