#   html2pml --- Helper Library to convert HTML Fragments to PML
#                (a markup language for pdf generation)
#
#   Copyright (C) 2003  net-labs Systemhaus GmbH
#                       info@net-labs.de
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with this library; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#
#   $Header: /cvsroot/collective/CMFReportTool/html2pml.py,v 1.8 2004/07/29 04:33:20 damonlynch Exp $
#   $Revision: 1.8 $



import re
from copy import copy
from types import StringType, DictType, TupleType, ListType, NoneType, UnicodeType
from xml.sax.writer import PrettyPrinter, XmlWriter
from cStringIO import StringIO


# python2.1 compatiblity
try:
    x = True
    del x
except:
    True = 1
    False = 0

      
# DISABLED --> NEED TO CLEANUP
def DEBUG(*args):
    #print ' '.join(args)
    pass


try:
    from mx.Tidy import Tidy
    HASTIDY = True
except ImportError:  
    HASTIDY = False

# Newstyle converter Constants

CONTENT_TAGS = ['para', 'image', 'table']
FORMAT_TAGS  = ['font', 'b', 'i', 'u', 'a', 'super', 'sub']
LISTING_TAGS = ['dl', 'ol', 'ul', 'li', 'dd', 'dt']
TABLE_TAGS   = ['tr', 'td', 'th']
PML_TAGS     = ['spacer', 'br']
ALLOWED_TAGS =  CONTENT_TAGS + FORMAT_TAGS + LISTING_TAGS + TABLE_TAGS + PML_TAGS

ALLOWED_ATTRIBUTES = {'para':    ['style'],
                      'image':   ['filename', 'height', 'width'],
                      'font':    ['color', 'face', 'size'],
                      'a':       ['href'],
                      }

ALLOWED_CHILDS = {'table':  ['tr'],
                  'tr':     ['td','th'],
                  'td':     ['para', 'image', '#text', 'table'],
                  'th':     ['para', 'image', '#text', 'table'],                              
                  'para':   ['ol', 'ul', 'dl', '#text', 'font', 'b', 'i', 'u', 'a', 'super', 'sub', 'br'],
                  'ol':     ['li'],
                  'ul':     ['li'],
                  'dl':     ['dt', 'dd'],
                  'li':     ['para', '#text', 'font', 'b', 'i', 'u', 'a', 'super', 'sub'],
                  'dt':     ['para', '#text', 'font', 'b', 'i', 'u', 'a', 'super', 'sub'],
                  'dd':     ['para', '#text', 'font', 'b', 'i', 'u', 'a', 'super', 'sub'],
                  'font':   ['#text', 'b', 'i', 'u', 'a', 'super', 'sub'],
                  'b':      ['#text', 'font', 'i', 'u', 'a', 'super', 'sub'],
                  'i':      ['#text', 'font', 'b', 'u', 'a', 'super', 'sub'],
                  'u':      ['#text', 'font', 'b', 'i', 'a', 'super', 'sub'],
                  'a':      ['#text', 'font', 'b', 'i', 'u', 'super', 'sub'],
                  'super':  ['#text', 'b', 'i', 'u', 'a', 'font'],
                  'sub':    ['#text', 'b', 'i', 'u', 'a', 'font']
                }


def __inverseDict(data):
    newDict = {}
    for k, v in data.items():
        if type(v) is ListType:
            for v1 in v:
                d = newDict.setdefault(v1, [])
                if not k in d and k != v1:
                    d.append(k)
        else:
            d = newDict.setdefault(v, [])
            if not k in d:
                d.append(k)
    return newDict

ALLOWED_PARENTS = __inverseDict(ALLOWED_CHILDS)


NEEDED_PARENT_TAG = {'ul':    'para',
                     'ol':    'para',
                     'dl':    'para',
                     'font':  'para',
                     'b':     'para',
                     'i':     'para',
                     'u':     'para',
                     'a':     'para',
                     'super': 'para',
                     'sub':   'para',
                     '#text': 'para',
                     }

TRANSFORM_TAG = { 'p':      'para',
                  'h1':     'para',
                  'h2':     'para',
                  'h3':     'para',
                  'h4':     'para',
                  'h5':     'para',
                  'h6':     'para',
                  'img':    'image',
                  'code':   'font',
                  'strong': 'b',
                  'em':      'i',
                  'hr':     'spacer',
                  'sup':    'super',
                  }


DEFAULT_ATTRIBUTES = {'para':  {'style': 'BodyText%(level)s'},
                      'spacer':{'height': '10'},
                      # XXX USING oldName as DictKey as well
                      # for keeping HeadingXX and the like !!!
                      'h1':    {'style': 'Heading1'},
                      'h2':    {'style': 'Heading2'},
                      'h3':    {'style': 'Heading3'},
                      'h4':    {'style': 'Heading4'},
                      'h5':    {'style': 'Heading4'},
                      'h6':    {'style': 'Heading4'},
                      'p':     {'style': 'BodyText%(level)s'},
                      'ul':    {'style': 'Bullet%(level)s',
                                'type' : 'square'},
                      'dl':    {'style': 'DefinitionList%(level)s',
                                'type' : ''},                                
                      'ol':    {'style': 'Sequence%(level)s'},
                      'table': {'style': 'StandardTable'},
                      'code':  {'face': 'Courier'},
                      }


UNWANTED_EMPTY_TAGS = ['para', 'table'] + FORMAT_TAGS + LISTING_TAGS + TABLE_TAGS


# Bug in Epoz Editor:
# scaling of images does not change
# height attribute correctly
regex_size_in_px = re.compile(r'^(?P<size>[0-9]+)px$').match

def transformAttributes_image(oldName, attributes, level):
    if 'style' in attributes.keys():
        for i in [n.strip() for n in attributes.get('style').split(';')]:
            key, val = i.split(':', 1)
            key = key.lower()
            if key in ['height','width']:
                match = regex_size_in_px(val.strip())
                if match:
                    attributes[key] = match.group('size')
    attributes['filename'] = attributes.get('src', '')

TRANSFORM_ATTRIBUTES = {'image': transformAttributes_image,
                       }




replacelist = [("&lt;", "<"), ("&gt;", ">"), ("&nbsp;", " "), ("&amp;", "&")] # , ("&quot;", '"')

def unEscapeContentList(contentList):
    result = []
    from string import replace
    for e in contentList:
        if "&" in e:
            for (old, new) in replacelist:
                e = replace(e, old, new)
        result.append(e)
    return result



class Html2Pml:

    _dispatch      = {}
    writerClass    = XmlWriter


    def convert(self, html, optimize=True, isXHTML=False):
        from reportlab.lib import rparsexml
        
        if not isXHTML:
            if not HASTIDY:
                raise ImportError, 'Could not import mx.Tidy'
            else:
                xhtml = self._convertToXhtml(html)
        else:
            if isinstance(html, UnicodeType):
                html = html.encode('utf-8')
            xhtml = html
        data  = rparsexml.parsexmlSimple(xhtml, entityReplacer=unEscapeContentList)

        program = []
        level   = 0
        self._do_convert(data, program, level)
        result = self._format(program, optimize=optimize)
        return result

    # handlers
    def _handle_none(self, data, program, level, info={}):
        pass

    def _handle_tag(self, tag, program, level, info={}):
        name, attributes, children, dummy = tag
        info = copy(info)
        DEBUG('handle_tag', name)

        undo = []
        todo = []
        convAttrs = self._convert_attributes
        
        newName = self._convert_tag(name, level)
       
        if not newName in ALLOWED_TAGS:
            self._do_convert(children, program, level)
        else:
            need_parentCheck = True
            # check if Tag is allowed in this context
            allowedTags = info.get('allowedTags', None)
            if allowedTags is not None:
                if not newName in allowedTags:
                    if newName in CONTENT_TAGS:
                        # create todo/undo code to flatten
                        todo, undo = self._program_flatten_ops(newName, program)
                else:
                    need_parentCheck = False
                del info['allowedTags']

            parentTag = info.get('parentTag', None)
            neededParent = NEEDED_PARENT_TAG.get(newName, None)
            if need_parentCheck and neededParent and neededParent != parentTag:
                todo.append(('startElement', neededParent, convAttrs(None, neededParent, {},  level), level))
                undo = [('endElement', neededParent, level,)] + undo

            info['parentTag'] = newName

            if todo:
                program.extend(todo)
            
            pline = ('startElement', newName, self._convert_attributes(name, newName, attributes, level), level,)
            program.append(pline)
            
            self._do_convert(children, program, level+1, info)

            # special case: don't render empty paragraphs
            if program[-1] == pline and pline[1] == 'para':
                program.pop()
            else:
                program.append(('endElement', newName, level,))

            if undo:
                program.extend(undo)

   
    def _handle_children(self, children, program, level, info={}):
        DEBUG('handle_children')
        info = copy(info)
        if level > 0:
            info['allowedTags'] = ALLOWED_CHILDS.get(info.get('parentTag', None), [])
        elif info.has_key('allowedTags'):
            del info['allowedTags']

        len_children = len(children)
        c = 0
        while c < len_children:
            child = children[c]
            if type(child) == StringType:
                while c+1 < len_children and type(children[c+1]) is StringType:
                    child += children[c+1]
                    c += 1
                
            self._do_convert(child, program, level, info)

            c += 1




    def _handle_text(self, text, program, level, info={}):
        text = text.strip()
        if not text:
            return

        DEBUG('handle_text', text)
        info = copy(info)
        todo = []
        undo = []
        convAttrs = self._convert_attributes

        need_parentCheck = True
        # check if Tag is allowed in this context
        allowedTags = info.get('allowedTags', None)
        if allowedTags is not None:
            if '#text' in allowedTags:
                need_parentCheck = False
            del info['allowedTags']

        parentTag = info.get('parentTag', None)
        neededParent = NEEDED_PARENT_TAG.get('#text', None)
        if need_parentCheck and neededParent and neededParent != parentTag:
            todo.append(('startElement', neededParent, convAttrs(None, neededParent, {},  level), level))
            undo = [('endElement', neededParent, level,)] + undo

        if todo:
            program.extend(todo)

        program.append(('characters', text, 0, len(text), level,))
                
        if undo:
            program.extend(undo)

            
    

    # configure dispatcher
    _dispatch[NoneType]   = _handle_none
    _dispatch[TupleType]  = _handle_tag
    _dispatch[ListType]   = _handle_children
    _dispatch[StringType] = _handle_text

    def _do_convert(self, data, program, level, info={}):
        DEBUG('do_convert', str(info.items()))
        self._dispatch[type(data)](self, data, program, level, info)



    # converters
    def _convert_tag(self, name, level):
        return TRANSFORM_TAG.get(name, name)


    def _convert_attributes(self, oldName, newName, attributes, level):
        DEBUG('convert_attributes', newName)
        if attributes is None:
            attributes = {}

        # transform attribute key/values
        func = TRANSFORM_ATTRIBUTES.get(newName, None)
        if func:
            func(oldName, attributes, level)
    
        # cleanup attributes
        allowedKeys = ALLOWED_ATTRIBUTES.get(newName, [])
        for key in attributes.keys():
            if not key in allowedKeys:
                del attributes[key]

        # setting default attributes
        if level and level > 0 and level <= 4:
            context = {'level': level}
        elif level > 4:
            context = {'level': 4}
        else:
            context = {'level': ''}
            
        defaults = DEFAULT_ATTRIBUTES.get(newName, {}).copy()
        # defaults for oldName overwrite defaults for newName !!!
        defaults.update(DEFAULT_ATTRIBUTES.get(oldName, {}))
        for key, value in defaults.items():
            if not attributes.has_key(key):
                attributes[key] = value % context

        return attributes

    
    # program helpers

    def _program_flatten_ops(self, tagName, program):
        todo = []
        undo = []
        count = 1
        # sensible default ??
        old_level = 99999
        len_program = len(program)

        allowedParents = ALLOWED_PARENTS.get(tagName, [])
        
        while count <= len_program:
            line = program[len_program - count]
            level = line[-1]
            opCode = line[0]
            
            if level > old_level:
                count += 1
                continue
            
            old_level = level

            if opCode == 'startElement':
                if line[1] in allowedParents:
                    break
                todo.append(('endElement', line[1], line[-1],))
                undo.append(line)

            if level == 0:
                break

            count += 1

        undo.reverse()
        return todo, undo

    def _optimize_program(self, program):
        result = []
        len_program = len(program)
        c = 0
        while c < len_program:
            line = program[c]
            opCode = line[0]
            # remove unwanted empty tags
            if opCode == 'startElement':
                tagName = line[1]
                if tagName in UNWANTED_EMPTY_TAGS:
                    nextLine = program[c+1]
                    nextOpCode = nextLine[0]
                    if nextOpCode == 'endElement' and \
                       nextLine[1] == tagName:
                        # empty unwanted tag,
                        # skip this and next line
                        c += 2
                        continue
                    
            result.append(line)
            c += 1
        
        return result



    # output
    def _format(self, program, optimize=True):
        fd = StringIO()
        output = self.writerClass(fd)
        if optimize:
            program = self._optimize_program(program)
        for line in program:
            getattr(output, line[0])(*line[1:-1])
        fd.seek(0)
        return unicode(fd.read(), 'utf-8')


    # external helpers    
    def _convertToXhtml(self, html, prepend_header=True, encoding='utf8'):
        nerrors, nwarnings, xhtml, errordata = Tidy.tidy( html.encode(encoding), output_xhtml = 1,
                                                              char_encoding = encoding)
        if prepend_header:
            xhtml = ('<?xml version="1.0" encoding="%s" ?>' % encoding) + xhtml
        return xhtml


